/*
    umicom.c - U-Boot stanalone app for handling Micom on Samsung A7/8/9 TV

    Copyright (C) 2013 Tomas Vanek

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "common.h"
#include "asm/io.h"
#include <exports.h>
#include "asm/stb7100reg.h"

#define BAUDMODE        0x00001000
#define CTSENABLE       0x00000800
#define FIFOENABLE      0x00000400
#define RXENABLE        0x00000100
#define RUN             0x00000080
#define LOOPBACK        0x00000000
#define STOPBIT         0x00000008
#define MODE            0x00000001
#define MODE_7BIT_PAR   0x0003
#define MODE_8BIT_PAR   0x0007
#define MODE_8BIT       0x0001
#define STOP_1BIT       0x0008
#define PARITYODD       0x0020

#define STA_NKD         0x0400
#define STA_TF          0x0200
#define STA_RHF         0x0100
#define STA_TOI         0x0080
#define STA_TNE         0x0040
#define STA_OE          0x0020
#define STA_FE          0x0010
#define STA_PE          0x0008
#define STA_THE         0x0004
#define STA_TE          0x0002
#define STA_RBF         0x0001

//#define TEST_ON_CONSOLE

#ifndef TEST_ON_CONSOLE
#undef CONFIG_SH_ASC_BASE
#define CONFIG_SH_ASC_BASE 0xb8030000
#endif

#define CONFIG_SH_ASC0_BASE 0xb8030000
#define CONFIG_SH_ASC2_BASE 0xb8032000

#define UART_BAUDRATE_OFFSET    0x00
#define UART_TXBUFFER_OFFSET    0x04
#define UART_RXBUFFER_OFFSET    0x08
#define UART_CONTROL_OFFSET     0x0C
#define UART_INTENABLE_OFFSET   0x10
#define UART_STATUS_OFFSET      0x14
#define UART_GUARDTIME_OFFSET   0x18
#define UART_TIMEOUT_OFFSET     0x1C
#define UART_TXRESET_OFFSET     0x20
#define UART_RXRESET_OFFSET     0x24
#define UART_RETRIES_OFFSET     0x28

#define UART_BAUDRATE_REG   (CONFIG_SH_ASC_BASE + UART_BAUDRATE_OFFSET)
#define UART_TXBUFFER_REG   (CONFIG_SH_ASC_BASE + UART_TXBUFFER_OFFSET)
#define UART_RXBUFFER_REG   (CONFIG_SH_ASC_BASE + UART_RXBUFFER_OFFSET)
#define UART_CONTROL_REG    (CONFIG_SH_ASC_BASE + UART_CONTROL_OFFSET)
#define UART_INTENABLE_REG  (CONFIG_SH_ASC_BASE + UART_INTENABLE_OFFSET)
#define UART_STATUS_REG     (CONFIG_SH_ASC_BASE + UART_STATUS_OFFSET)
#define UART_GUARDTIME_REG  (CONFIG_SH_ASC_BASE + UART_GUARDTIME_OFFSET)
#define UART_TIMEOUT_REG    (CONFIG_SH_ASC_BASE + UART_TIMEOUT_OFFSET)
#define UART_TXRESET_REG    (CONFIG_SH_ASC_BASE + UART_TXRESET_OFFSET)
#define UART_RXRESET_REG    (CONFIG_SH_ASC_BASE + UART_RXRESET_OFFSET)
#define UART_RETRIES_REG    (CONFIG_SH_ASC_BASE + UART_RETRIES_OFFSET)


/*---- Values for the BAUDRATE Register -----------------------*/


#if defined(CONFIG_SH_STB7100)
#define PCLK                    (gd->bd->bi_emifrq*1000000)
#define BAUDRATE_VAL_M0(bps)    (PCLK / (16 * (bps)))
#define BAUDRATE_VAL_M1(bps)    ( (((bps * (1 << 14))+ (1<<13)) / (PCLK/(1 << 6))))
#else
#define PCLK                    B_CLOCK_RATE
#define BAUDRATE_VAL_M0(bps)    (PCLK / (16 * (bps)))
#define BAUDRATE_VAL_M1(bps)    (int)((((double)bps * (1 << 20))/ PCLK)+0.5)
#endif

/*
 * MODE 0
 *                       ICCLK
 * ASCBaudRate =   ----------------
 *                   baudrate * 16
 *
 * MODE 1
 *                   baudrate * 16 * 2^16
 * ASCBaudRate =   ------------------------
 *                          ICCLK
 */

#define BPS_2400        BAUDRATE_VAL_M0(2400)
#define BPS_4800        BAUDRATE_VAL_M0(4800)
#define BPS_9600        BAUDRATE_VAL_M0(9600)
#define BPS_19200       BAUDRATE_VAL_M1(19200)
#define BPS_38400       BAUDRATE_VAL_M1(38400)
#define BPS_57600       BAUDRATE_VAL_M1(57600)
#define BPS_96000       BAUDRATE_VAL_M1(96000)
#define BPS_115200      BAUDRATE_VAL_M1(115200)


#if defined(CONFIG_STM_ST231)

#define p2_outl(addr,b) writel(b,addr)
#define p2_inl(addr)    readl(addr)

#endif

//////////////////////////////////////////////////////////////////////////////

#ifdef UIMAGE_XIP
unsigned long uimage_hdr[64/4] __attribute__ ((section (".init"), used)) = {
  0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
  0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
  0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
  0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff,
};
#endif
int umicom(int argc, char* argv[]);// __attribute__ ((section (".text")));
int CheckGd(void) __attribute__ ((section (".text2")));
void xputc(char c) __attribute__ ((section (".text2")));
void xputs(const char *s) __attribute__ ((section (".text2")));
int atoi(const char *p) __attribute__ ((section (".text2")));
void MicomSendByte(unsigned char d) __attribute__ ((section (".text2")));
void MicomSend6(unsigned char* bf) __attribute__ ((section (".text2")));
void MicomSend2(int cmd, int p1) __attribute__ ((section (".text2")));
void SetClockgenAPll1(unsigned long cfg) __attribute__ ((section (".text2")));
void SetBaudRate(void) __attribute__ ((section (".text2"), nothrow));
void MemTest(void) __attribute__ ((section (".text2")));
void MemTestBlock(unsigned long beg, unsigned long end) __attribute__ ((section (".text2")));

unsigned long __udivsi3_i4i(unsigned long, unsigned long);

int gd_err;

//////////////////////////////////////////////////////////////////////////////
/*
 U-Boot commands

 nfs 84300000 $nfsserver:$nfsrootdir/boot/umicom.img;go 84300040
     load and run app, switches off watchdogs

 go 84300040 ovc ndiv mdiv pexp
 go 84300040 ovc 67 9 0
    overclock pll1 to 402MHz
*/

int umicom(int argc, char* argv[])
{
  app_startup(argv);

  gd_err= CheckGd();

  if (gd_err) {
    xputs("U-Boot global data block not found");
    xputs("use command 'go addr'");
  } else {
    gd->flags&= ~GD_FLG_SILENT;
  }
  if (gd_err || argc < 2) {
    xputs("=== www.SamyGO.tv === umicom.img 1.1 === micom watchdog killer ===");
    xputs("http://wiki.samygo.tv/index.php5?title=Watchdogs_and_MICOM_on_A_series(SH4)");
    if (gd_err) {
      xputs("micom via AsyncSerialController0");
    } else {
      printf("micom via AsyncSerialController0, 0x%x\n", UART_BAUDRATE_REG);
    }
    xputs(" baud rate set to 9600 Bd");
    p2_outl (UART_CONTROL_REG, MODE_8BIT | STOP_1BIT | RUN);
    SetBaudRate();
    MicomSend2(208, 0);
    xputs(" ...msg 'Set Sw Pvcc' sent");
    MicomSend2(232, 1);
    xputs(" ...msg 'App Init Ok' sent");
    if (gd_err) {
      MicomSend2(22, 0);  // proper return to U-Boot is uncertain
      xputs(" ...msg 'Watchdog Enable' sent");
      MicomSend2(227, 1);
      xputs(" ...power LED slowly blinking");
    } else {
      MicomSend2(23, 0);
      xputs(" ...msg 'Watchdog Disable' sent");
      MicomSend2(227, 3);
      xputs(" ...power LED quickly blinking");
    }
    if (! gd_err) {
      printf("Use:\n"
             " go %x e  ....enable watchdog\n", umicom);
      printf(" go %x d  ....disable watchdog\n", umicom);
      printf(" go %x 18  ...switch off the TV\n", umicom);
      printf(" go %x 29  ...power cycle reboot\n", umicom);
      printf(" go %x 143  ..reset\n", umicom);
      printf(" go %x ram  ..presence test of RAM\n", umicom);
      printf(" go %x ovc [ndiv [mdiv [pexp]]] ..overclock ClkGenA PLL1\n", umicom);
    }
    return 0;
  }

  if (argc >= 2) {
    if (argv[1][0] == 'o' && argv[1][1] == 'v' && argv[1][2] == 'c') {
      int cfg, ndiv= 30, mdiv= 2, pexp= 1;
      if (argc >= 3) ndiv= atoi(argv[2]) & 0xff;
      if (argc >= 4) mdiv= atoi(argv[3]) & 0xff;
      if (argc >= 5) pexp= atoi(argv[4]) & 7;
      cfg= 0x80000 | (pexp << 16) | (ndiv << 8) | mdiv;

      MicomSend2(227, 1);
      xputs(" ...power LED slowly blinking");
      MicomSend2(22, 0);
      xputs(" ...msg 'Watchdog Enable' sent\n");

      printf("New pll1 cfg 0x%x\n", cfg);
      udelay(2000);  // wait for UART to finish transmit

      SetClockgenAPll1(cfg);
//      MicomSend2(23, 0);
      return 0;
    }
  }

  if (argc == 2) {
    if (argv[1][0] == 'r' && argv[1][1] == 'a' && argv[1][2] == 'm') {
      MemTest();
      return 0;
    } else if (argv[1][0] == 'e' || argv[1][0] == 'E') {
      MicomSend2(22, 0);
      xputs(" ...msg 'Watchdog Enable' sent");
      MicomSend2(227, 1);
      xputs(" ...power LED slowly blinking");
    } else if (argv[1][0] == 'd' || argv[1][0] == 'D') {
      MicomSend2(23, 0);
      xputs(" ...msg 'Watchdog Disable' sent");
      MicomSend2(227, 3);
      xputs(" ...power LED quickly blinking");
    } else {
      MicomSend2(atoi(argv[1]), 0);
      xputs(" ...msg sent\n");
    }
    return 1;
  }

  return 0;
}

//////////////////////////////////////////////////////////////////////////////

int CheckGd(void)
{
//  printf("gd %08x\n", gd);
  if (! gd) {
    return -1;
  }
//  printf("flg %x  baud %d  haveCons %x  ram_sz %x  reloc_off %x  env_addr %x  jt %x\n",
//     gd->flags, gd->baudrate, gd->have_console, gd->ram_size, gd->reloc_off, gd->env_addr, (int)(gd->jt));
//  printf("bd %x  baudrate %d\n", gd->bd, gd->bd->bi_baudrate);
  if (! gd->env_addr || (unsigned int)(gd->env_addr) - (unsigned int)gd > 0x90000
    || ! gd->jt || (unsigned int)(gd->jt) - (unsigned int)gd > 0x90000
  ) {
    return -2;
  }
  if (! gd->bd || (unsigned int)gd - (unsigned int)(gd->bd) > 0x1000 || gd->baudrate != gd->bd->bi_baudrate) {
    return -3;
  }
  return 0;
}

//////////////////////////////////////////////////////////////////////////////

void xputc(char c)
{
  unsigned long status;

  do {
    status = p2_inl (CONFIG_SH_ASC2_BASE + UART_STATUS_OFFSET);
  } while (status & STA_TF);

  p2_outl (CONFIG_SH_ASC2_BASE + UART_TXBUFFER_OFFSET, c);
}

void xputs(const char *s)
{
  if (gd_err) {
    while (*s) xputc(*s++);
    xputc('\r');
    xputc('\n');
  } else {
    printf("%s\n", s);
  }
}

int atoi(const char *p)
{
  int r= 0;
  while (1) {
    if (*p >= '0' && *p <= '9') {
      r= 10 * r + *p++ - '0';
    } else {
      return r;
    }
  }
}

void MicomSendByte(unsigned char d)
{
  unsigned long status;

  do {
    status = p2_inl (UART_STATUS_REG);
  } while (status & STA_TF);

#ifndef TEST_ON_CONSOLE
  if (! gd_err) {
    printf(" %02x", d);
  }
#endif
  p2_outl (UART_TXBUFFER_REG, d);
}

void MicomSend6(unsigned char* bf)
{
  int i, sum;

  MicomSendByte(0xff);
  MicomSendByte(0xff);
  for (i= 0, sum= 0; i < 6; i++) {
    MicomSendByte(bf[i]);
    sum+= bf[i];
  }
  MicomSendByte(sum);
}

void MicomSend2(int cmd, int p1)
{
  unsigned char bf[6];
  bf[0]= (unsigned char)cmd;
  bf[1]= (unsigned char)p1;
  bf[2]= 0;
  bf[3]= 0;
  bf[4]= 0;
  bf[5]= 0;
  MicomSend6(bf);
}

//////////////////////////////////////////////////////////////////////////////

void SetBaudRate(void)
{
  unsigned long pll, mdiv, ndiv, pdiv, pll1freq;
  unsigned long brdiv, bps, q;
  pll= *STB7100_CLOCKGENA_PLL1_CFG;
  mdiv= pll & 0xff;
  if (mdiv == 0) mdiv= 1;
  ndiv= (pll >> 8) & 0xff;
  pdiv= (pll >> 16) & 0x7;

  // use of integer div (__udivsi3_i4i) break code relocatibility - GOT does not get relocated
  //pll1freq = ((2000 * INPUT_CLOCK_RATE * ndiv)/ mdiv) / (1 << pdiv);
  pll1freq = __udivsi3_i4i(2000 * INPUT_CLOCK_RATE * ndiv, mdiv << pdiv);
  bps= 115200;  // ExLink console, BaudMode 1
  q= 1000 * pll1freq / (4 * 4 * 16);

  // use of integer div (__udivsi3_i4i) break code relocatibility - GOT does not get relocated
  // brdiv= ((bps << 14) + q / 2) / q;
  brdiv= __udivsi3_i4i((bps << 14) + q / 2, q);
  p2_outl(CONFIG_SH_ASC2_BASE + UART_BAUDRATE_OFFSET, brdiv);
  printf("pll1 config is 0x%x (%d kHz), serial console baud divisor %d\n", pll, pll1freq, brdiv);

  bps= 9600;  // micom, BaudMode 0
  q= (4 * 16) * bps;
  brdiv= (1000 * pll1freq + q / 2) / q;
  p2_outl(CONFIG_SH_ASC0_BASE + UART_BAUDRATE_OFFSET, brdiv);
  printf("baud divisor for micom %d\n", brdiv);

  gd->bd->bi_pll1frq= pll1freq / 1000;
  gd->bd->bi_st231frq = gd->bd->bi_pll1frq;
  gd->bd->bi_stbusfrq = gd->bd->bi_pll1frq/2;
  gd->bd->bi_lmifrq = gd->bd->bi_pll1frq/2;
  gd->bd->bi_emifrq = gd->bd->bi_pll1frq/4;
}

//////////////////////////////////////////////////////////////////////////////

void SetClockgenAPll1(unsigned long cfg)
{
  unsigned long pll, byp;
  int i= 0, j;

  pll= *STB7100_CLOCKGENA_PLL1_CFG & ~0x80000;
  byp= *STB7100_CLOCKGENA_PLL1_BYPASS & ~2;
  do {
    // first round (i= 0) of the loop just reads program to the cache
    // registers are locked, no changes take place
    // second round (i= 1) do real changes
    *STB7100_CLOCKGENA_LOCK= i;
    *STB7100_CLOCKGENA_PLL1_BYPASS= byp | 2;  // switch clk to bypass
    *STB7100_CLOCKGENA_PLL1_CFG= pll; // pll1 off
    *STB7100_CLOCKGENA_PLL1_CFG= cfg; // pll1 set and switch on
    while (! (*STB7100_CLOCKGENA_PLL1_STATUS & 1)); // wait for pll1 lock
    *STB7100_CLOCKGENA_PLL1_BYPASS= byp;  // bypass off
    *STB7100_CLOCKGENA_LOCK= 0; // lock registers
//    udelay(2000);  // wait for clock to stabilize
    j= 0;
    do {
      asm volatile("");
      j++;
    } while (j < i);  // wait for clock to stabilize
    i+= 0xc0de;
  } while (i <= 0xc0de);

  SetBaudRate();
}

//////////////////////////////////////////////////////////////////////////////

#define LMI_SYS_START 0xA4000000
#define LMI_SYS_END   0xAF000000
#define LMI_VID_START 0xB0000000
#define LMI_VID_END   0xB7000000
#define PAGE  4096

void MemTest(void)
{
  MemTestBlock(LMI_VID_START, LMI_VID_END);
  MemTestBlock(LMI_SYS_START, LMI_SYS_END);
}

void MemTestBlock(unsigned long beg, unsigned long end)
{
  unsigned long addr;
  unsigned long lastGoodAddr= 0;
  int mirrors= 0;
  int badPages= 0;
  int pattern1= 0x55555555;
  int pattern2= 0xaaaaaaaa;
  for (addr= beg; addr < end; addr+= PAGE) {
    if (addr >= 0xA4300000 && addr < 0xA4700000) {
      continue; // skip this code and U-Boot
    }
    unsigned long *p= (void*)addr;
    if (p[0] == 0 && p[1] == pattern1
        && p[2] == pattern2 && p[3] == 0xffffffff
        && p[4] == 0 && p[6] == 0
        && p[5] + p[7] == 0xffffffff) {
      if (addr == beg) {
        pattern1= 0xaaaaaaaa;
        pattern2= 0x55555555;
      } else {
        printf("Found mirror of %08lx at %08lx\n", p[5], addr);
        if (++mirrors > 8) break;
        continue;
      }
    }
    lastGoodAddr= addr;
    if ((addr & 0xffffff) == 0) {
      printf("Writing %08lx\n", addr);
    }
    p[0]= 0;
    p[1]= pattern1;
    p[2]= pattern2;
    p[3]= 0xffffffff;
    p[4]= 0;
    p[5]= addr;
    p[6]= 0;
    p[7]= 0xffffffff - addr;
  }
  end= lastGoodAddr;
  for (addr= beg; addr <= end; addr+= PAGE) {
    if (addr >= 0xA4300000 && addr < 0xA4700000) {
      continue; // skip this code and U-Boot
    }
    unsigned long *p= (void*)addr;
    if (p[0] == 0 && p[1] == pattern1
        && p[2] == pattern2 && p[3] == 0xffffffff
        && p[4] == 0 && p[6] == 0
        && p[5] + p[7] == 0xffffffff) {
      // memory retention ok
      if ((addr & 0xffffff) == 0) {
        printf("Testing %08lx\n", addr);
      }
      if (p[5] == addr) {
        lastGoodAddr= addr;
      } else {
        printf("Found mirror of %08lx at %08lx\n", p[5], addr);
        if (++mirrors > 8) break;
        continue;
      }
    } else {
      printf("Bad data at %08lx\n", addr);
      if (++badPages > 8) break;
    }
  }
  unsigned long kib= (lastGoodAddr + PAGE - beg) / 1024;
  printf("Memory ends at %08lx, size is %d kiB or %d MiB\n", lastGoodAddr + PAGE - 1, kib, kib / 1024);
}
